using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Text;
using System.Reflection;
using Microsoft.Ink;

namespace Handwriting_Identification
{
    static class Program
    {
        public class stroke
        {
            public int averagePressure;
            public int endPressure;
            public int duration;
            public ArrayList points;

            public stroke (int aPres, int ePres, int dur, ArrayList pts){
                averagePressure = aPres;
                endPressure = ePres;
                duration = dur;
                points = pts;
            }

            public string ToString()
            {
                String rtrn;
                Point[] allpoints = (Point[])points.ToArray(typeof(Point));
                rtrn = "Avg. Pressure:" + averagePressure + ", End Pressure:" + endPressure + ", Duration:" + duration + " [" + allpoints[0].ToString();
                for (int i = 1; i < allpoints.Length; i++)
                {
                    rtrn += ", " + allpoints[i].ToString();
                }
                rtrn += "]";

                return rtrn;
            }
        }

        public class letter
        {
            public ArrayList strokes;
            public char thisChar;
            public char prevChar;
            public char nextChar;

            public letter(char tc, char pc, char nc, ArrayList s)
            {
                strokes = s;
                thisChar = tc;
                prevChar = pc;
                nextChar = nc;
            }

            public string ToString()
            {
                String rtrn;
                stroke [] allstrokes = (stroke [])strokes.ToArray(typeof(stroke));
                rtrn = "Chars:" + prevChar + ", " + thisChar.ToString() + ", " + nextChar.ToString() + Environment.NewLine;
                for (int i = 0; i < allstrokes.Length; i++)
                {
                    rtrn += "     " + allstrokes[i].ToString() + Environment.NewLine;
                }

                return rtrn;
            }
        }

        public class Form1 : Form
        {
            private static Recognizers recognizers = new Recognizers();
            private Panel m_pnlInput;
            private Label m_labelReco;
            private Label m_labelSupported;
            private Label m_labelProperites;
            private InkOverlay m_inkOverlay;
            private TextBox m_textStrokes;
            private Button m_btnOutput;
            private int totalPressure;
            private int totalSamples;
            private int totalTime;
            private String prevRec;
            private ArrayList chars;
            private ArrayList curStrokes;
            private Strokes newStrokes;
            private letter curLetter;
            private string fullString;
            private int curIndex;

            int colorT;

            static void Main()
            {
                Application.Run(new Form1());
            }

            public Form1()
            {
                Control.CheckForIllegalCrossThreadCalls = false;

                SuspendLayout();

                m_pnlInput = new Panel();
                m_pnlInput.BorderStyle = BorderStyle.Fixed3D;
                m_pnlInput.BackColor = Color.White;
                m_pnlInput.Location = new Point(8, 8);
                m_pnlInput.Size = new Size(1000, 150);
                m_pnlInput.Paint += new PaintEventHandler(drawGrid);

                m_btnOutput = new Button();
                m_btnOutput.Location = new Point(8, 204);
                m_btnOutput.Size = new Size(100, 24);
                m_btnOutput.Text = "Output";
                m_btnOutput.Click += new EventHandler(recognize);

                m_labelReco = new Label();
                m_labelReco.Location = new Point(112, 204);
                m_labelReco.Size = new Size(896, 24);

                m_textStrokes = new TextBox();
                m_textStrokes.Multiline = true;
                m_textStrokes.AcceptsReturn = true;
                m_textStrokes.AcceptsTab = true;
                m_textStrokes.ReadOnly = true;
                m_textStrokes.ScrollBars = ScrollBars.Both;
                m_textStrokes.Location = new Point(8, 232);
                m_textStrokes.Size = new Size(1000, 200);

                m_labelSupported = new Label();
                m_labelSupported.Location = new Point(8, 436);
                m_labelSupported.Size = new Size(1000, 24);
                m_labelSupported.Text = "Properties supported: ";

                m_labelProperites = new Label();
                m_labelProperites.Location = new Point(8, 464);
                m_labelProperites.Size = new Size(1000, 16);

                AutoScaleBaseSize = new Size(5, 13);
                ClientSize = new Size(1018, 488);
                Controls.AddRange(new Control[] { m_pnlInput, m_labelReco, m_labelSupported, m_textStrokes, m_labelProperites, m_btnOutput });
                FormBorderStyle = FormBorderStyle.FixedDialog;
                MaximizeBox = false;
                Text = "Handwriting Identification";

                ResumeLayout(false);

                m_inkOverlay = new InkOverlay(m_pnlInput.Handle);
                m_inkOverlay.EraserMode = InkOverlayEraserMode.PointErase;
                m_inkOverlay.EraserWidth = 200;

                Tablet t = (new Tablets()).DefaultTablet;
                m_inkOverlay.SetSingleTabletIntegratedMode(t);

                foreach (FieldInfo f in typeof(PacketProperty).GetFields())
                {
                    if(f.IsStatic && f.IsPublic)
                    {
                        Guid g = (Guid)f.GetValue(f);
                        if (t.IsPacketPropertySupported(g))
                        {
                            m_labelSupported.Text += f.Name + ", ";
                        }
                    }
                }

                Guid [] desiredProps = new Guid[4];
                desiredProps[0] = PacketProperty.X;
                desiredProps[1] = PacketProperty.Y;
                desiredProps[2] = PacketProperty.NormalPressure;
                desiredProps[3] = PacketProperty.TimerTick;
                m_inkOverlay.DesiredPacketDescription = desiredProps;
                m_inkOverlay.DefaultDrawingAttributes.Color = Color.Black;
          
                m_inkOverlay.CursorDown += new InkCollectorCursorDownEventHandler(handleCursorDown);
                m_inkOverlay.Stroke += new InkCollectorStrokeEventHandler(handleStroke);
                m_inkOverlay.NewPackets += new InkCollectorNewPacketsEventHandler(handleNewPackets);

                prevRec = "";
                curStrokes = new ArrayList();
                chars = new ArrayList();
                colorT = 1;

                curIndex = 0;
                fullString = "A B C D E F G H I J K L M N O P Q R S T U V W X Y Z a b c d e f g h i j k l m n o p q r s t u v w x y z The quick brown fox jumps over the lazy dog. How razorback-jumping frogs can level six piqued gymnasts! Back in June we delivered oxygen equipment of the same size.";

                m_inkOverlay.Enabled = true;
            }

            void drawGrid (object sender,PaintEventArgs e)
            {
                Graphics g = e.Graphics;

                int width = m_pnlInput.Width / 100;
                int height = m_pnlInput.Height;
                Font tFont = new Font(FontFamily.GenericSerif, 9);
                g.DrawString(fullString.Substring(curIndex, 1), tFont, Brushes.White, 0, 0);
                for (int i = 1; i < 100; i++)
                {
                    if (curIndex + i < fullString.Length) g.DrawString(fullString.Substring(curIndex + i, 1), tFont, Brushes.White, i * width + 2, 0);
                    g.DrawLine(Pens.LightGray, i * width, 0, i * width, height);
                }
                for (int j = 1; j <= 16; j++)
                {
                    g.DrawLine(Pens.LightGray, 0, j*(height / 16), width * 100, j*(height / 16));
                }
            }

            void handleCursorDown(object sender, InkCollectorCursorDownEventArgs e)
            {
                totalPressure = 0;
                totalSamples = 0;
                totalTime = 0;
            }

            void handleStroke(object sender, InkCollectorStrokeEventArgs e)
            {
                String curRec;

                Recognizer recognizer = Form1.recognizers.GetDefaultRecognizer();
                RecognizerContext recoCtxt = recognizer.CreateRecognizerContext();

                recoCtxt.CharacterAutoCompletion = RecognizerCharacterAutoCompletionMode.Full;

                recoCtxt.StopBackgroundRecognition();
                recoCtxt.EndInkInput();

                recoCtxt.Strokes = m_inkOverlay.Ink.Strokes;

                RecognitionStatus recoStatus;
                RecognitionResult recoResult = recoCtxt.Recognize(out recoStatus);

                curRec = recoResult.TopString;
                curRec = curRec.Replace(" ", "");

                m_labelReco.Text = curRec + " : " + prevRec;

                int endPressure = e.Stroke.GetPacketValuesByProperty(PacketProperty.NormalPressure)[0];

                ArrayList pts = new ArrayList();
                for (int i = 0; i < e.Stroke.BezierCusps.Length; i++)
                {
                    Point pt = new Point(e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].X, e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].Y);
                    pts.Add(pt);
                }

                stroke curStroke = new stroke((totalPressure / totalSamples), endPressure, totalTime, pts);

                if (curRec.Length > prevRec.Length && prevRec.Length > 0)
                {
                    curLetter = new letter(' ', ' ', ' ', curStrokes);
                    chars.Add(curLetter);

                    Graphics g = m_pnlInput.CreateGraphics();
                  
                    foreach (stroke tmpStroke in (stroke[])curLetter.strokes.ToArray(typeof(stroke)))
                    {
                        Point[] tmpPts = (Point[])tmpStroke.points.ToArray(typeof(Point));
                        m_inkOverlay.Renderer.InkSpaceToPixel(g, ref tmpPts);

                        foreach (Point pt in tmpPts)
                        {
                            Font tFont = new Font(FontFamily.GenericSansSerif, 9);
                            if (colorT % 2 == 0)
                            {
                                g.FillEllipse(Brushes.Crimson, pt.X - 2, pt.Y - 2, 5, 5);
                                g.DrawEllipse(Pens.Black, pt.X - 2, pt.Y - 2, 5, 5);
                            }
                            else
                            {
                                g.FillEllipse(Brushes.Blue, pt.X - 2, pt.Y - 2, 5, 5);
                                g.DrawEllipse(Pens.Black, pt.X - 2, pt.Y - 2, 5, 5);
                            }
                        }

                    }

                    colorT++;
                    curStrokes = new ArrayList();
                    
                }
               
                prevRec = curRec;
                curStrokes.Add(curStroke);

                recoCtxt.BackgroundRecognize();
            }

            private void handleNewPackets(object sender, InkCollectorNewPacketsEventArgs e)
            {
                /*String str = "";
                for (int i = 0; i < e.PacketData.Length; i++)
                {
                    str += e.PacketData[i].ToString() + ", ";
                }
                m_labelProperites.Text = str;*/

                totalPressure += e.PacketData[2];
                totalTime = e.PacketData[3];
                totalSamples++;
            }

            private void recognize (object sender, EventArgs e)
            {
                Recognizer recognizer = Form1.recognizers.GetDefaultRecognizer();
                RecognizerContext recoCtxt = recognizer.CreateRecognizerContext();

                recoCtxt.CharacterAutoCompletion = RecognizerCharacterAutoCompletionMode.Full;
                

                recoCtxt.StopBackgroundRecognition();
                recoCtxt.Strokes = m_inkOverlay.Ink.Strokes;

                RecognitionStatus recoStatus;
                RecognitionResult recoResult = recoCtxt.Recognize(out recoStatus);

                recoCtxt.EndInkInput();
                
                output(recoResult.TopString, (letter[])chars.ToArray(typeof(letter)));
            }

            private void output(String txt, letter [] letterList)
            {
                //int offset = 0;
                //char [] stripped = prevRec.Replace(" ", "").ToCharArray();
                String curRec = " " + txt  + " ";
                char [] orig = curRec.ToCharArray();
                int j = 0;
                for (int i = 1; i < (orig.Length - 1); i++)
                {
                    if (orig[i] != ' ' && j < letterList.Length){
                        letterList[j].prevChar = orig[i - 1];
                        letterList[j].thisChar = orig[i];
                        letterList[j].nextChar = orig[i + 1];
                        j++;
                    }
                }
                m_textStrokes.Text = "";
                for (int i = 0; i < letterList.Length; i++)
                {
                    m_textStrokes.Text = letterList[i].ToString() + m_textStrokes.Text;
                }
            }

        }
    }
}