using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Text;
using System.IO;
using System.Reflection;
using Microsoft.Ink;
using SVM;

namespace Handwriting_Identification
{
	public class hIdentifier : Form
	{
		private static Recognizers recognizers = new Recognizers();
		private Panel m_pnlInput;
		private Label m_labelReco;
		private Label m_labelSupported;
		private Label m_labelProperites;
		private InkOverlay m_inkOverlay;
		private TextBox m_textStrokes;
		private Button m_btnOutput;
		private int totalPressure;
		private int totalSamples;
		private int totalTime;
		private String prevRec;
		private ArrayList chars;
		private ArrayList curStrokes;
        private ArrayList curPressures;
		private Strokes newStrokes;
		private letter curLetter;
		private string fullString;
		private int curIndex;
        private int curUser;
        private int maxStrokes;
        private FileStream fileStream;

		int colorT;

		public class stroke
		{
			public int averagePressure;
            public int endPressure;
			public int duration;
            public int[] inkGrid;
            public ArrayList pressures;
            public ArrayList points;

			public stroke(int aPres, int ePres, int dur, ArrayList pts)
			{
				averagePressure = aPres;
				endPressure = ePres;
				duration = dur;
				points = pts;
                inkGrid[200] = new int();
                for (int i = 0; i < 200; i++) inkGrid[i] = 0;
			}

            public string ToString()
			{
				String rtrn;
				Point[] allpoints = (Point[])points.ToArray(typeof(Point));
				rtrn = "Avg. Pressure:" + averagePressure + ", End Pressure:" + endPressure + ", Duration:" + duration + " [" + allpoints[0].ToString();
				for (int i = 1; i < allpoints.Length; i++)
				{
					rtrn += ", " + allpoints[i].ToString();
				}
				rtrn += "]";

				return rtrn;
			}
		}

		public class letter
		{
            public ArrayList strokes;
            public Boolean[] inkGrid;
            public char thisChar;
            public char prevChar;
            public char nextChar;

			public letter(char tc, char pc, char nc, ArrayList s)
			{
				strokes = s;
				thisChar = tc;
				prevChar = pc;
				nextChar = nc;
			}

            public void addStroke(stroke s){
                strokes.Add(strokes);
            }

			public string ToString()
			{
				String rtrn;
				stroke[] allstrokes = (stroke[])strokes.ToArray(typeof(stroke));
				rtrn = "Chars:" + prevChar + ", " + thisChar.ToString() + ", " + nextChar.ToString() + Environment.NewLine;
				for (int i = 0; i < allstrokes.Length; i++)
				{
					rtrn += "     " + allstrokes[i].ToString() + Environment.NewLine;
				}

				return rtrn;
			}
		}

        public hIdentifier(String collectText, int user, FileStream fs)
		{
			Control.CheckForIllegalCrossThreadCalls = false;

			SuspendLayout();

			m_pnlInput = new Panel();
			m_pnlInput.BorderStyle = BorderStyle.Fixed3D;
			m_pnlInput.BackColor = Color.DarkGray;
			m_pnlInput.Location = new Point(8, 8);
			m_pnlInput.Size = new Size(1000, 80);
			m_pnlInput.Paint += new PaintEventHandler(drawGrid);

			m_btnOutput = new Button();
			m_btnOutput.Location = new Point(907, 93);
			m_btnOutput.Size = new Size(100, 20);
			m_btnOutput.Text = "Next";
			m_btnOutput.Click += new EventHandler(process);

			m_labelReco = new Label();
			m_labelReco.Location = new Point(112, 194);
			m_labelReco.Size = new Size(896, 24);

			m_textStrokes = new TextBox();
			m_textStrokes.Multiline = true;
			m_textStrokes.AcceptsReturn = true;
			m_textStrokes.AcceptsTab = true;
			m_textStrokes.ReadOnly = true;
			m_textStrokes.ScrollBars = ScrollBars.Both;
			m_textStrokes.Location = new Point(8, 119);
			m_textStrokes.Size = new Size(1000, 216);

			/*m_labelSupported = new Label();
			m_labelSupported.Location = new Point(8, 436);
			m_labelSupported.Size = new Size(1000, 24);
			m_labelSupported.Text = "Properties supported: ";

			m_labelProperites = new Label();
			m_labelProperites.Location = new Point(8, 450);
			m_labelProperites.Size = new Size(1000, 16);*/

			AutoScaleBaseSize = new Size(5, 13);
			ClientSize = new Size(1017, 344);
			Controls.AddRange(new Control[] {m_pnlInput, m_textStrokes, m_btnOutput});
			FormBorderStyle = FormBorderStyle.FixedDialog;
			MaximizeBox = false;
			Text = "Handwriting Identification";

			ResumeLayout(false);

			m_inkOverlay = new InkOverlay(m_pnlInput.Handle);
			m_inkOverlay.EraserMode = InkOverlayEraserMode.PointErase;
			m_inkOverlay.EraserWidth = 200;

			Tablet t = (new Tablets()).DefaultTablet;
			m_inkOverlay.SetSingleTabletIntegratedMode(t);

			/*foreach (FieldInfo f in typeof(PacketProperty).GetFields())
			{
				if(f.IsStatic && f.IsPublic)
				{
					Guid g = (Guid)f.GetValue(f);
					if (t.IsPacketPropertySupported(g))
					{
						m_labelSupported.Text += f.Name + ", ";
					}
				}
			}*/

			Guid [] desiredProps = new Guid[4];
			desiredProps[0] = PacketProperty.X;
			desiredProps[1] = PacketProperty.Y;
			desiredProps[2] = PacketProperty.NormalPressure;
			desiredProps[3] = PacketProperty.TimerTick;
			m_inkOverlay.DesiredPacketDescription = desiredProps;
			m_inkOverlay.DefaultDrawingAttributes.Color = Color.White;
	  
			m_inkOverlay.CursorDown += new InkCollectorCursorDownEventHandler(handleCursorDown);
			m_inkOverlay.Stroke += new InkCollectorStrokeEventHandler(handleStroke);
			m_inkOverlay.NewPackets += new InkCollectorNewPacketsEventHandler(handleNewPackets);

			prevRec = "";
			curStrokes = new ArrayList();
			chars = new ArrayList();
            curPressures = new ArrayList();
			colorT = 1;

			curIndex = 1;
            fullString = " " + collectText + " ";
            curUser = user;
            fileStream = fs;
            maxStrokes = 0;
                
            //" A B C D E F G H I J K L M N O P Q R S T U V W X Y Z a b c d e f g h i j k l m n o p q r s t u v w x y z The quick brown fox jumps over the lazy dog How razorback jumping frogs can level six piqued gymnasts Back in June we delivered oxygen equipment of the same size To be or not to be that is the question Whether tis nobler in the mind to suffer The slings and arrows of outrageous fortune ";
			//fullString = " Sixty zippers were quickly picked from the woven jute bag Brick quiz whangs jumpy veldt fox This is a test ";

			m_inkOverlay.Enabled = true;
		}

		void drawGrid (object sender,PaintEventArgs e)
		{
			Graphics g = e.Graphics;
			
			int width = m_pnlInput.Width/25;
			int height = m_pnlInput.Height;
			Font tFont = new Font(FontFamily.GenericSerif, 9);
			g.DrawString(fullString.Substring(curIndex, 1), tFont, Brushes.White, 0, 0);
			for (int i = 1; i <= 25; i++)
			{
				if (curIndex + i < fullString.Length) g.DrawString(fullString.Substring(curIndex + i, 1), tFont, Brushes.White, i * width + 2, 0);
				g.DrawLine(Pens.White, i * width, 0, i * width, height);
			}
			
		}

		void handleCursorDown(object sender, InkCollectorCursorDownEventArgs e)
		{
			totalPressure = 0;
			totalSamples = 0;
			totalTime = 0;
		}

		void handleStroke(object sender, InkCollectorStrokeEventArgs e)
		{
			int endPressure = e.Stroke.GetPacketValuesByProperty(PacketProperty.NormalPressure)[0];

			ArrayList pts = new ArrayList();
			for (int i = 0; i < e.Stroke.BezierCusps.Length; i++)
			{
				Point pt = new Point(e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].X, e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].Y);
				pts.Add(pt);
			}

			stroke curStroke = new stroke((totalPressure / totalSamples), endPressure, totalTime, pts);
			curStrokes.Add(curStroke);
		}

		private void handleNewPackets(object sender, InkCollectorNewPacketsEventArgs e)
		{
			/*String str = "";
			for (int i = 0; i < e.PacketData.Length; i++)
			{
				str += e.PacketData[i].ToString() + ", ";
			}
			m_labelProperites.Text = str;*/

			totalPressure += e.PacketData[2];
            curPressures.Add(e.PacketData[2]);
			totalTime = e.PacketData[3];
			totalSamples++;
		}

        private void outputCSV()
        {
            StreamWriter sw = new StreamWriter(fileStream);
            sw.Write("PrevChar,ThisChar,NextChar,NumStrokes,Width,Height,Duration,AveragePressure");
            for (int i = 0; i < 5; i++)
            {
                sw.Write(",Stroke" + i.ToString() + "Duration,Stroke" + i.ToString() + "AveragePressure,Stroke" + i.ToString() + "EndPressure");
                for (int j = 0; j < 10; j++)
                {
                    sw.Write(",Stroke" + i.ToString() + "Point" + j.ToString() + "X,Stroke" + i.ToString() + "Point" + j.ToString() + "Y");
                }
            }
            sw.Write(",User" + Environment.NewLine);
            foreach (letter l in chars)
            {
                sw.Write(l.prevChar + "," + l.thisChar + "," + l.nextChar + "," + l.strokes.Count + ",");
                int left = 2000;
                int right = -100;
                int top = 2000;
                int bottom = -100;
                int dur = 0;
                int totalPressure = 0;
                stroke[] allS = (stroke[])l.strokes.ToArray(typeof(stroke));
                foreach (stroke s in allS)
                {
                    Point[] pts = (Point[])s.points.ToArray(typeof(Point));
                    for (int i = 0; i < pts.Length; i++)
                    {
                        if (pts[i].X < left) left = pts[i].X;
                        if (pts[i].X > right) right = pts[i].X;
                        if (pts[i].Y < top) top = pts[i].Y;
                        if (pts[i].Y > bottom) bottom = pts[i].Y;
                    }
                    dur += s.duration;
                    totalPressure += s.duration * s.averagePressure;
                }
                sw.Write((right - left).ToString() + "," + (bottom - top).ToString() + "," + dur + "," + (totalPressure / dur).ToString());
                for (int i = 0; i < 5; i++)
                {
                    if (i < allS.Length)
                    {
                        sw.Write("," + allS[i].duration + "," + allS[i].averagePressure + "," + allS[i].endPressure);
                        Point[] pts = (Point[])allS[i].points.ToArray(typeof(Point));
                        for (int j = 0; j < 10; j++)
                        {
                            if (j < pts.Length)
                                sw.Write("," + pts[j].X.ToString() + "," + pts[j].Y.ToString());
                            else
                                sw.Write(",0,0");
                        }
                    }
                    else
                    {
                        sw.Write(",0,0,0");
                        for (int j = 0; j < 10; j++) sw.Write(",0,0");
                    }
                }
                sw.Write("," + curUser + Environment.NewLine);
            }
            sw.Close();
        }

        private void outputSVM()
        {
            StreamWriter sw = new StreamWriter(fileStream);
            foreach (letter l in chars)
            {
                sw.Write(curUser +
                         " 1:" + ((Convert.ToInt32(l.prevChar) + 1.0) / 256.0) +
                         " 2:" + ((Convert.ToInt32(l.thisChar) + 1.0) / 256.0) +
                         " 3:" + ((Convert.ToInt32(l.nextChar) + 1.0) / 256.0) +
                         " 4:" + ((l.strokes.Count * 1.0) / 6.0));
                int left = 2000;
                int right = -100;
                int top = 2000;
                int bottom = -100;
                int dur = 0;
                int totalPressure = 0;
                stroke[] allS = (stroke[])l.strokes.ToArray(typeof(stroke));
                foreach (stroke s in allS)
                {
                    Point[] pts = (Point[])s.points.ToArray(typeof(Point));
                    for (int i = 0; i < pts.Length; i++)
                    {
                        if (pts[i].X < left) left = pts[i].X;
                        if (pts[i].X > right) right = pts[i].X;
                        if (pts[i].Y < top) top = pts[i].Y;
                        if (pts[i].Y > bottom) bottom = pts[i].Y;
                    }
                    dur += s.duration;
                    totalPressure += s.duration * s.averagePressure;
                }
                sw.Write(" 5:" + (right - left) / (m_pnlInput.Width / 25.0) +
                         " 6:" + (bottom - top) / 90.0 +
                         " 7:" + (dur / 3000.0) +
                         " 8:" + (totalPressure / dur) / 1000.0);
                for (int i = 0; i < 5; i++)
                {
                    if (i < allS.Length)
                    {
                        sw.Write(" " + (9 + (i * 23)) + ":" + allS[i].duration / 1000.0 +
                                 " " + (10 + (i * 23)) + ":" + allS[i].averagePressure / 256.0 +
                                 " " + (11 + (i * 23)) + ":" + allS[i].endPressure / 256.0);
                        Point[] pts = (Point[])allS[i].points.ToArray(typeof(Point));
                        for (int j = 0; j < 10; j++)
                        {
                            if (j < pts.Length)
                                sw.Write(" " + (12 + (i * 23) + (j * 2)) + ":" + pts[j].X / (m_pnlInput.Width / 25.0) +
                                         " " + (13 + (i * 23) + (j * 2)) + ":" + pts[j].Y / 90.0);
                        }
                    }
                }
                sw.Write(Environment.NewLine);
            }
            sw.Close();
            fileStream.Close();
        }

		private void process(object sender, EventArgs e)
		{
			if (curIndex < fullString.Length)
			{
				Graphics g = m_pnlInput.CreateGraphics();
				letter[] tmpCharList = new letter[25];
				for (int i = 0; i <= 25; i++)
				{
					if (curIndex + i + 1 < fullString.Length)
						tmpCharList[i] = new letter(fullString.ToCharArray()[curIndex + i], fullString.ToCharArray()[curIndex + i - 1], fullString.ToCharArray()[curIndex + i + 1], new ArrayList());
					else
						tmpCharList[i] = new letter(' ', ' ', ' ', new ArrayList());
				}
				int width = m_pnlInput.Width / 25;

				foreach (stroke s in curStrokes.ToArray())
				{
					Point[] pts = (Point[])s.points.ToArray(typeof(Point));
					m_inkOverlay.Renderer.InkSpaceToPixel(g, ref pts);
					int left = pts[0].X;
					int right = pts[0].X;
					int top = pts[0].Y;
					int bottom = pts[0].Y;
					for (int i = 1; i < pts.Length; i++)
					{
						if (pts[i].X < left) left = pts[i].X;
						if (pts[i].X > right) right = pts[i].X;
						if (pts[i].Y < top) top = pts[i].Y;
						if (pts[i].Y > bottom) bottom = pts[i].Y;
					}
					int half = ((right - left) / 2) + left;
					int cIndex = 0;
					for (int i = 1; i <= 25; i++)
                        if (half > (width * i)) cIndex = i;
					for (int i = 0; i < pts.Length; i++) pts[i].X -= cIndex * width;
					s.points = new ArrayList();
					for (int i = 0; i < pts.Length; i++) s.points.Add(pts[i]);
					tmpCharList[cIndex].addStroke(s);
				}

				for (int i = 0; i < tmpCharList.Length; i++)
				{
					if (tmpCharList[i].strokes.Count != 0)
					{
						chars.Add(tmpCharList[i]);
                        if (tmpCharList[0].strokes.Count > maxStrokes) maxStrokes = tmpCharList[0].strokes.Count;
						m_textStrokes.Text = tmpCharList[i].ToString() + m_textStrokes.Text;
					}
				}
				curIndex += 25;
				curStrokes = new ArrayList();
				m_inkOverlay.Ink.DeleteStrokes();
				m_pnlInput.Refresh();

				if (curIndex > fullString.Length)
				{
                    outputCSV();
                    m_btnOutput.Enabled = false;
				}
			}
		}
	}
}