using System;
using System.Drawing;
using System.Collections;
using System.Windows.Forms;
using System.Text;
using System.IO;
using System.Reflection;
using Microsoft.Ink;
using SVM;

namespace Handwriting_Identification
{
	public class hIdentifier : Form
	{
        //Global Variables
		private static Recognizers recognizers = new Recognizers();
		private Panel m_pnlInput;
		private Label m_labelReco;
		//private Label m_labelSupported;
		//private Label m_labelProperites;
		private InkOverlay m_inkOverlay;
		private TextBox m_textStrokes;
		private Button m_btnOutput;
        private Graphics pnlInputGfx;

        private FileStream fileStream;
        private StreamWriter sw;

        private stroke curStroke;
        private ArrayList strokes;
        private int curIndex;
        private int totalTime;
        private string fullString;
        private int curUser;
        private char[] fullStringArray;

        //stroke structure, contains collection of points to form a stroke
		public class stroke
		{
            public bool[] grid;
            public ArrayList pressures;
            public double[] pressureStats;
            public int duration;
            public int[] boundries;
            public ArrayList breziers;
            public ArrayList points;
            public int cell;

            public stroke()
            {
                //initialize everything
                grid = new bool[200];
                for (int i = 0; i < 200; i++) grid[i] = false;
                pressures = new ArrayList();
                points = new ArrayList();
                pressureStats = new double[7];
                duration = 0;
                boundries = new int[4];
                boundries[0] = 1000; //min x
                boundries[1] = 80; //min y
                boundries[2] = 0; //max x
                boundries[3] = 0; //max y
                breziers = new ArrayList();
                cell = 0;
            }

            public void addPacket(int x, int y, int p)
            {
                //set boundries
                if (x < boundries[0]) boundries[0] = x; //set current min x
                if (y < boundries[1]) boundries[1] = y; //set current min y
                if (x > boundries[2]) boundries[2] = x; //set current max x
                if (y > boundries[3]) boundries[3] = y; //set current max y

                //add the new point
                points.Add(new Point(x,y));

                //add current pressure
                pressures.Add(p);
            }

            private void writeToGrid()
            {
                int curCell;
                foreach (Point p in (Point[])points.ToArray(typeof(Point)))
                {
                    curCell = (10 * (p.Y / 10)) + (p.X / 10);
                    if (curCell <= 199 || curCell >= 0) grid[curCell] = true;
                }
            }

            private void calcStats()
            {
                //find center
                int centerX = (boundries[2] + boundries[0]) / 2;
                cell = centerX / 40;

                //normalize all points to that of the first cell
                boundries[0] = boundries[0] - (40 * cell);
                boundries[2] = boundries[2] - (40 * cell);
                ArrayList convertedPoints = new ArrayList();
                foreach (Point p in (Point[])points.ToArray(typeof(Point))) 
                    convertedPoints.Add(new Point(p.X - (40 * cell), p.Y));
                points = convertedPoints;

                //write to grid
                writeToGrid();

                //calculate pressure statistics
                int[] pArray = (int[])pressures.ToArray(typeof(int));
                int total = 0;
                int min = pArray[0];
                int max = pArray[0];
                double avg;
                //get total pressure, min, max
                foreach (int p in pArray)
                {
                    total += p;
                    if (p > max) max = p;
                    if (p < min) min = p;
                }
                //get average pressure
                avg = (double)total/(double)pArray.Length;
                //calculate standard deviation
                double sumPow = 0;
                double stdDev;
                foreach (int p in pArray) sumPow += Math.Pow(avg - p, 2);
                stdDev = Math.Sqrt((1.0 / (double)pArray.Length) * (double)sumPow);
                //set the stats
                pressureStats[0] = pArray[0];
                pressureStats[1] = pArray[pArray.Length - 1];
                pressureStats[2] = min;
                pressureStats[3] = max;
                pressureStats[4] = avg;
                pressureStats[5] = total;
                pressureStats[6] = stdDev;
            }

            public void finalize(ArrayList b, int d)
            {
                //finalize stroke
                breziers = b;
                duration = d;
                calcStats();
            }

            public string ToString()
            {
                string returnStr = "";
                returnStr = "Stroke: (" + boundries[0] + ", " + boundries[1] + "), (" + boundries[2] + ", " + boundries[3] + ") of cell " + cell + Environment.NewLine;
                returnStr += "     Pressure Statistics: " + pressureStats[2] + "(min) " + pressureStats[3] + "(max) "
                                    + pressureStats[4] + "(avg) " + pressureStats[5] + "(sum) " + pressureStats[6] + "(stddev)" + Environment.NewLine;
                return returnStr;
            }
		}

        //letter structure, contains collection of strokes to form a letter
		public class letter
		{
            public int[] grid;
            public double[] pressureStats;
            public char[] letters;
            public int[] boundries;
            public ArrayList breziers;
            public ArrayList pressures;
            public int numStrokes;
            public int[] size;
            public int duration;

			public letter(char pc, char tc, char nc)
			{
                //initialize everything
				grid = new int[200];
                for (int i = 0; i < 200; i++) grid[i] = 0;
                pressureStats = new double[5];
                letters = new char[3];
                letters[0] = pc;
                letters[1] = tc;
                letters[2] = nc;
                breziers = new ArrayList();
                pressures = new ArrayList();
                boundries = new int[4];
                boundries[0] = 40; //min x
                boundries[1] = 80; //min y
                boundries[2] = 0; //max x
                boundries[3] = 0; //max y
                size = new int[2];
                size[0] = 0; //width
                size[1] = 0; //height
                numStrokes = 0;
                duration = 0;
			}

            public void addStroke(stroke s)
            {
                //update total duration
                duration += s.duration;
                //update number of strokes
                numStrokes++;
                //update boundries
                if (s.boundries[0] < boundries[0]) boundries[0] = s.boundries[0];
                if (s.boundries[1] < boundries[1]) boundries[1] = s.boundries[1];
                if (s.boundries[2] > boundries[2]) boundries[2] = s.boundries[2];
                if (s.boundries[3] > boundries[3]) boundries[3] = s.boundries[3];
                //update size
                size[0] = boundries[2] - boundries[0];
                size[1] = boundries[3] - boundries[1];
                //update grid
                for (int i = 0; i < 200; i++) if (s.grid[i]) grid[i]++;
                //add breziers
                breziers.AddRange(s.breziers);
                //edit pressureStats
                pressures.AddRange(s.pressures);
                int[] pArray = (int[])pressures.ToArray(typeof(int));
                int max = pArray[0];
                int min = pArray[0];
                int sum = 0;
                double sumPow = 0.0;
                double stdDev;
                foreach (int p in pArray)
                {
                    if (p > max) max = p;
                    if (p < min) min = p;
                    sum += p;
                }
                double avg = (double)sum / (double)pArray.Length;
                foreach (int p in pArray) sumPow += Math.Pow(avg - p, 2);
                stdDev = Math.Sqrt((1.0 / (double)pArray.Length) * (double)sumPow);
                pressureStats[0] = min;
                pressureStats[1] = max;
                pressureStats[2] = avg;
                pressureStats[3] = sum;
                pressureStats[4] = stdDev;
            }

            public string ToString()
            {
                string returnStr = "";
                returnStr = "Character: (" + letters[0] + ") " + letters[1] + " (" + letters[2] + ") | (" 
                                    + numStrokes + " strokes)" + Environment.NewLine;
                returnStr += "     Pressure Statistics: " + pressureStats[0] + "(min) " + pressureStats[1] + "(max) "
                                    + pressureStats[2] + "(avg) " + pressureStats[3] + "(sum) " + pressureStats[4] + "(stddev)" + Environment.NewLine;
                returnStr += "     Boundry: (" + boundries[0] + ", " + boundries[1] + "), (" + boundries[2] + ", " + boundries[3] + ")" + Environment.NewLine;
                returnStr += "     Size: " + size[0] + "x" + size[1] + Environment.NewLine;
                returnStr += "     Duration: " + duration + Environment.NewLine;
                return returnStr;
            }
        }

        //create form
        public hIdentifier(String collectText, int user, FileStream fs)
		{
			Control.CheckForIllegalCrossThreadCalls = false;

			SuspendLayout();

			m_pnlInput = new Panel();
			m_pnlInput.BorderStyle = BorderStyle.Fixed3D;
			m_pnlInput.BackColor = Color.DarkGray;
			m_pnlInput.Location = new Point(8, 8);
			m_pnlInput.Size = new Size(1000, 80);
			m_pnlInput.Paint += new PaintEventHandler(drawGrid);

			m_btnOutput = new Button();
			m_btnOutput.Location = new Point(907, 93);
			m_btnOutput.Size = new Size(100, 20);
			m_btnOutput.Text = "Next";
			m_btnOutput.Click += new EventHandler(process);

			m_labelReco = new Label();
			m_labelReco.Location = new Point(112, 194);
			m_labelReco.Size = new Size(896, 24);

			m_textStrokes = new TextBox();
			m_textStrokes.Multiline = true;
			m_textStrokes.AcceptsReturn = true;
			m_textStrokes.AcceptsTab = true;
			m_textStrokes.ReadOnly = true;
			m_textStrokes.ScrollBars = ScrollBars.Both;
			m_textStrokes.Location = new Point(8, 119);
			m_textStrokes.Size = new Size(1000, 216);

			/*m_labelSupported = new Label();
			m_labelSupported.Location = new Point(8, 436);
			m_labelSupported.Size = new Size(1000, 24);
			m_labelSupported.Text = "Properties supported: ";

			m_labelProperites = new Label();
			m_labelProperites.Location = new Point(8, 450);
			m_labelProperites.Size = new Size(1000, 16);*/

			AutoScaleBaseSize = new Size(5, 13);
			ClientSize = new Size(1017, 344);
			Controls.AddRange(new Control[] {m_pnlInput, m_textStrokes, m_btnOutput});
			FormBorderStyle = FormBorderStyle.FixedDialog;
			MaximizeBox = false;
			Text = "Handwriting Identification";

            this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.hIdentifier_Close);

			ResumeLayout(false);

			m_inkOverlay = new InkOverlay(m_pnlInput.Handle);
			m_inkOverlay.EraserMode = InkOverlayEraserMode.PointErase;
			m_inkOverlay.EraserWidth = 200;

			Tablet t = (new Tablets()).DefaultTablet;
			m_inkOverlay.SetSingleTabletIntegratedMode(t);

			/*foreach (FieldInfo f in typeof(PacketProperty).GetFields())
			{
				if(f.IsStatic && f.IsPublic)
				{
					Guid g = (Guid)f.GetValue(f);
					if (t.IsPacketPropertySupported(g))
					{
						m_labelSupported.Text += f.Name + ", ";
					}
				}
			}*/

			Guid [] desiredProps = new Guid[4];
			desiredProps[0] = PacketProperty.X;
			desiredProps[1] = PacketProperty.Y;
			desiredProps[2] = PacketProperty.NormalPressure;
			desiredProps[3] = PacketProperty.TimerTick;
			m_inkOverlay.DesiredPacketDescription = desiredProps;
			m_inkOverlay.DefaultDrawingAttributes.Color = Color.White;
	  
			m_inkOverlay.CursorDown += new InkCollectorCursorDownEventHandler(handleCursorDown);
			m_inkOverlay.Stroke += new InkCollectorStrokeEventHandler(handleStroke);
			m_inkOverlay.NewPackets += new InkCollectorNewPacketsEventHandler(handleNewPackets);

            //initialize global variables
			curIndex = 1;
            fullString = " " + collectText + " ";
            curUser = user;
            strokes = new ArrayList();
            fullStringArray = fullString.ToCharArray();
            pnlInputGfx = m_pnlInput.CreateGraphics();
            //initialize file outputs
            fileStream = fs;
            sw = new StreamWriter(fileStream);
            initCSV();
                
            //" A B C D E F G H I J K L M N O P Q R S T U V W X Y Z a b c d e f g h i j k l m n o p q r s t u v w x y z The quick brown fox jumps over the lazy dog How razorback jumping frogs can level six piqued gymnasts Back in June we delivered oxygen equipment of the same size To be or not to be that is the question Whether tis nobler in the mind to suffer The slings and arrows of outrageous fortune ";
			//fullString = " Sixty zippers were quickly picked from the woven jute bag Brick quiz whangs jumpy veldt fox This is a test ";

			m_inkOverlay.Enabled = true;
		}

		void drawGrid (object sender,PaintEventArgs e)
		{
            //draw grid for data collection
			Graphics g = e.Graphics;
			
			int width = m_pnlInput.Width/25;
			int height = m_pnlInput.Height;
			Font tFont = new Font(FontFamily.GenericSerif, 9);
			for (int i = 0; i < 25; i++)
			{
				if (curIndex + i < fullString.Length - 1) g.DrawString(fullStringArray[i + curIndex].ToString(), tFont, Brushes.White, i * width + 2, 0);
                g.DrawLine(Pens.White, (i + 1) * width, 0, (i + 1) * width, height);
			}
			
		}

        //initialize a new stroke
		void handleCursorDown(object sender, InkCollectorCursorDownEventArgs e)
		{
            //initialize a new stroke struct
            curStroke = new stroke();
			totalTime = 0;
		}

        //end a new stroke
		void handleStroke(object sender, InkCollectorStrokeEventArgs e)
		{
            //obtain brezier cusps
			ArrayList breziers = new ArrayList();
			for (int i = 0; i < e.Stroke.BezierCusps.Length; i++)
			{
				Point pt = new Point(e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].X, e.Stroke.BezierPoints[e.Stroke.BezierCusps[i]].Y);
                breziers.Add(pt);
			}
            //finalize stroke
            curStroke.finalize(breziers, totalTime);
			strokes.Add(curStroke);

            //display stroke stats
            //m_textStrokes.Text = curStroke.ToString() + m_textStrokes.Text;
		}

        //handles a new data packet
		private void handleNewPackets(object sender, InkCollectorNewPacketsEventArgs e)
		{
            //convert coordinates to relative to panal
            Point curPoint = new Point(e.PacketData[0], e.PacketData[1]);
            m_inkOverlay.Renderer.InkSpaceToPixel(pnlInputGfx, ref curPoint);
            //add new packets to stroke struct
            curStroke.addPacket(curPoint.X, curPoint.Y, e.PacketData[2]);
			totalTime = e.PacketData[3];
		}

        //initializes csv data file
        private void initCSV()
        {
            //write header
            sw.Write("PrevChar,ThisChar,NextChar,MinX,MinY,MaxX,MaxY,NumStrokes,Width,Height,Duration,MinPressure,MaxPressure,AvgPressure,TotPressure,StdDevPressure,NumBreziers");
            for (int i = 0; i < 15; i++) sw.Write(",BrezierX" + i.ToString() + ",BrezierY" + i.ToString());
            for (int i = 0; i < 200; i++) sw.Write(",Grid" + i.ToString());
            sw.Write(",User" + Environment.NewLine);
        }

        //output csv data file
        private void outputCSV(letter[] letters)
        {
            foreach (letter l in letters)
            {
                //write character information
                for (int i = 0; i < 3; i++) sw.Write(l.letters[i].ToString() + ",");
                //write boundry information
                for (int i = 0; i < 4; i++) sw.Write(l.boundries[i].ToString() + ",");
                //write other info
                sw.Write(l.numStrokes.ToString() + "," + l.size[0].ToString() + "," + l.size[1].ToString() + "," + l.duration.ToString() + ",");
                //write pressure info
                for (int i = 0; i < 5; i++) sw.Write(l.pressureStats[i].ToString() + ",");
                //write brezier info
                sw.Write(l.breziers.Count.ToString() + ",");
                Point[] b = (Point[])l.breziers.ToArray(typeof(Point));
                for (int i = 0; i < 15; i++)
                {
                    if (i < b.Length)
                        sw.Write(b[i].X.ToString() + "," + b[i].Y.ToString() + ",");
                    else
                        sw.Write("0,0,");
                }
                //write grid info
                for (int i = 0; i < 200; i++) sw.Write(l.grid[i].ToString() + ",");
                //write user and newline
                sw.Write(curUser.ToString() + Environment.NewLine);
            }
        }

        //output file for built in SVM
        private void outputSVM()
        {
            /*StreamWriter sw = new StreamWriter(fileStream);
            foreach (letter l in chars)
            {
                sw.Write(curUser +
                         " 1:" + ((Convert.ToInt32(l.prevChar) + 1.0) / 256.0) +
                         " 2:" + ((Convert.ToInt32(l.thisChar) + 1.0) / 256.0) +
                         " 3:" + ((Convert.ToInt32(l.nextChar) + 1.0) / 256.0) +
                         " 4:" + ((l.strokes.Count * 1.0) / 6.0));
                int left = 2000;
                int right = -100;
                int top = 2000;
                int bottom = -100;
                int dur = 0;
                int totalPressure = 0;
                stroke[] allS = (stroke[])l.strokes.ToArray(typeof(stroke));
                foreach (stroke s in allS)
                {
                    Point[] pts = (Point[])s.points.ToArray(typeof(Point));
                    for (int i = 0; i < pts.Length; i++)
                    {
                        if (pts[i].X < left) left = pts[i].X;
                        if (pts[i].X > right) right = pts[i].X;
                        if (pts[i].Y < top) top = pts[i].Y;
                        if (pts[i].Y > bottom) bottom = pts[i].Y;
                    }
                    dur += s.duration;
                    totalPressure += s.duration * s.averagePressure;
                }
                sw.Write(" 5:" + (right - left) / (m_pnlInput.Width / 25.0) +
                         " 6:" + (bottom - top) / 90.0 +
                         " 7:" + (dur / 3000.0) +
                         " 8:" + (totalPressure / dur) / 1000.0);
                for (int i = 0; i < 5; i++)
                {
                    if (i < allS.Length)
                    {
                        sw.Write(" " + (9 + (i * 23)) + ":" + allS[i].duration / 1000.0 +
                                 " " + (10 + (i * 23)) + ":" + allS[i].averagePressure / 256.0 +
                                 " " + (11 + (i * 23)) + ":" + allS[i].endPressure / 256.0);
                        Point[] pts = (Point[])allS[i].points.ToArray(typeof(Point));
                        for (int j = 0; j < 10; j++)
                        {
                            if (j < pts.Length)
                                sw.Write(" " + (12 + (i * 23) + (j * 2)) + ":" + pts[j].X / (m_pnlInput.Width / 25.0) +
                                         " " + (13 + (i * 23) + (j * 2)) + ":" + pts[j].Y / 90.0);
                        }
                    }
                }
                sw.Write(Environment.NewLine);
            }
            sw.Close();
            fileStream.Close();*/
        }

		private void process(object sender, EventArgs e)
		{
			if (curIndex < fullString.Length)
			{
                //calculate how many letters are being collected
                int numLetters = 25;
                if (fullString.Length - curIndex - 1 < 25) numLetters = fullString.Length - curIndex - 1;
                letter[] curLetters = new letter[numLetters];
                //initialize new letter structures to store and sort strokes
                for (int i = curIndex; i < curIndex + numLetters; i++)
                    curLetters[i-curIndex] = new letter(fullStringArray[i-1], fullStringArray[i], fullStringArray[i+1]);
                //sort strokes
                foreach(stroke s in (stroke[])strokes.ToArray(typeof(stroke)))
                    curLetters[s.cell].addStroke(s);
                //output strokes
                foreach(letter l in curLetters)
                    m_textStrokes.Text = l.ToString() + m_textStrokes.Text;

                //move to next set of letters
                curIndex += 25;
                strokes = new ArrayList();
                m_inkOverlay.Ink.DeleteStrokes();
                m_pnlInput.Refresh();

                //output to csv
                outputCSV(curLetters);
			}
		}

        //if closed without finishing, close file stream
        private void hIdentifier_Close(object sender, EventArgs e)
        {
            sw.Close();
            fileStream.Close();
        }
	}
}