using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Collections;
using System.Windows.Forms;
using System.IO;
using SVM;

namespace Handwriting_Identification
{
    public partial class Configure : Form
    {
        // Training Objects ///////////////////////////////
        private ArrayList trainSubLbls;
        private ArrayList trainSubBtns;
        private FileStream fileTrain;

        // Test Objects ///////////////////////////////////
        private ArrayList testSubLbls;
        private ArrayList testSubBtns;
        private FileStream fileTest;

        public Configure()
        {
            InitializeComponent();
        }

        static void Main()
        {
            Application.Run(new Configure());
        }

        private void TrainSubject(object sender, EventArgs e)
        {
            int curTrain = Convert.ToInt32(sender.ToString().Substring(sender.ToString().Length-1));
            fileTrain = new FileStream(filename.Text + curTrain.ToString() + ".train.csv", FileMode.Create, FileAccess.Write);
            hIdentifier newForm = new hIdentifier(trainTxt.Text, curTrain, fileTrain);
            newForm.Show();
            ((Label)(trainSubLbls.ToArray())[curTrain]).Text = "User " + curTrain.ToString() + ":Trained";
            ((Button)(trainSubBtns.ToArray())[curTrain]).Enabled = false;
        }

        private void TestSubject(object sender, EventArgs e)
        {
            int curTest = Convert.ToInt32(sender.ToString().Substring(sender.ToString().Length - 1));
            fileTest = new FileStream(filename.Text + curTest.ToString() + ".test.csv", FileMode.Create, FileAccess.Write);
           
            hIdentifier newForm = new hIdentifier(testTxt.Text, curTest, fileTest);
            newForm.Show();
            ((Label)(testSubLbls.ToArray())[curTest]).Text = "User " + curTest.ToString() + ":Test data recorded";
            ((Button)(testSubBtns.ToArray())[curTest]).Enabled = false;
        }

        private void initBtn_Click(object sender, EventArgs e)
        {
            int numberOfSubjects = 0;
            try
            {
                numberOfSubjects = Convert.ToInt32(numSubj.Text);
            }
            catch (Exception ex)
            {
                numberOfSubjects = -1;
            }

            if (numberOfSubjects > 0)
            {
                // Initialize training tab ///////////////////////////////////////////////////////////////
                trainSubLbls = new ArrayList();
                trainSubBtns = new ArrayList();

                for (int i = 0; i < numberOfSubjects; i++)
                {
                    Label tmpLabel = new Label();
                    tmpLabel.Location = new System.Drawing.Point(7, 7 + (29 * i));
                    tmpLabel.Name = "train" + i.ToString() + "Lbl";
                    tmpLabel.Size = new System.Drawing.Size(187, 23);
                    tmpLabel.TabIndex = 0;
                    tmpLabel.Text = "User " + i.ToString() + ": Not Complete";
                    tmpLabel.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;

                    Button tmpButton = new Button();
                    tmpButton.Location = new System.Drawing.Point(200, 7 + (29 * i));
                    tmpButton.Name = "train" + i.ToString() + "Btn";
                    tmpButton.Size = new System.Drawing.Size(75, 23);
                    tmpButton.TabIndex = 2;
                    tmpButton.Text = "Train U" + i.ToString();
                    tmpButton.UseVisualStyleBackColor = true;
                    tmpButton.Click += new System.EventHandler(TrainSubject);

                    trainSubLbls.Add(tmpLabel);
                    trainSubBtns.Add(tmpButton);

                    TrainTab.Controls.Add(tmpLabel);
                    TrainTab.Controls.Add(tmpButton);
                }

                // Initialize testing tab ////////////////////////////////////////////////////////////////
                testSubLbls = new ArrayList();
                testSubBtns = new ArrayList();

                fileTest = new FileStream(filename.Text + ".test", FileMode.Create, FileAccess.Write);
                fileTest.Close();

                for (int i = 0; i < numberOfSubjects; i++)
                {
                    Label tmpLabel = new Label();
                    tmpLabel.Location = new System.Drawing.Point(7, 7 + (29 * i));
                    tmpLabel.Name = "test" + i.ToString() + "Lbl";
                    tmpLabel.Size = new System.Drawing.Size(187, 23);
                    tmpLabel.TabIndex = 0;
                    tmpLabel.Text = "User " + i.ToString() + ": Not Complete";
                    tmpLabel.TextAlign = System.Drawing.ContentAlignment.MiddleLeft;

                    Button tmpButton = new Button();
                    tmpButton.Location = new System.Drawing.Point(200, 7 + (29 * i));
                    tmpButton.Name = "test" + i.ToString() + "Btn";
                    tmpButton.Size = new System.Drawing.Size(75, 23);
                    tmpButton.TabIndex = 2;
                    tmpButton.Text = "Test U" + i.ToString();
                    tmpButton.UseVisualStyleBackColor = true;
                    tmpButton.Click += new System.EventHandler(TestSubject);

                    testSubLbls.Add(tmpLabel);
                    testSubBtns.Add(tmpButton);

                    TestTab.Controls.Add(tmpLabel);
                    TestTab.Controls.Add(tmpButton);
                }
            }

            // Disable initialize button /////////////////////////////////////////////
            initBtn.Enabled = false;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            // First, read in the training data
            Problem train = Problem.Read("a1a.train"); 
            // Now, determine the range of the data so we can 
            // scale it to be uniform across all dimensions 
            Range range = Scaling.DetermineRange(train); 
            // We now scale all data using that range 
            train = Scaling.Scale(train, range); 
            Problem test = Problem.Read("a1a.test"); 
            test = Scaling.Scale(test, range); 
            // For this example (and indeed, many scenarios), the default 
            // parameters will suffice. 
            Parameter parameters = new Parameter();
            double C; 
            double Gamma; 
            // This will do a grid optimization to find the best parameters 
            // and store them in C and Gamma, outputting the entire 
            // search to params.txt. 
            ParameterSelection.Grid( train, parameters, "params.txt", out C, out Gamma ); 
            parameters.C = C; parameters.Gamma = Gamma; 
            // Train the model using the optimal parameters. 
            Model model = Training.Train(train, parameters); 
            // Perform classification on the test data, putting the 
            // results in results.txt. 
            Prediction.Predict(test, "results.txt", model, false);
        }
        //letter structure
        public class letterInst
        {
            public char letter;
            //grid 0-199
            public int[] grid;
            //size width, height
            public int[] size;
            //pressure average
            public int pressures;
            //duration average
            public int duration;
            //boundries x0, y0, x1, y1
            public int[] boundries;
            //private variables
            private int numInstances;
            private int[] sizeTotal;
            private int pressureTotal;
            private int durationTotal;
            //initialize
            public letterInst(char l)
            {
                letter = l;
                grid = new int[200];
                size = new int[2];
                pressures = 0;
                boundries = new int[4];
                boundries[0] = 2000;
                boundries[1] = 2000;

                numInstances = 0;
                sizeTotal = new int[2];
            }
            //add instance of letter
            public void add(int[] g, int[] s, int p, int d, int[] b)
            {
                numInstances++;
                //update grid
                for (int i = 0; i < 200; i++) grid[i] += g[i];
                //update size
                sizeTotal[0] += s[0];
                sizeTotal[1] += s[1];
                size[0] = sizeTotal[0] / numInstances;
                size[1] = sizeTotal[1] / numInstances;
                //update pressures
                pressureTotal += p;
                pressures = pressureTotal / numInstances;
                //update durations
                durationTotal += d;
                duration = durationTotal / numInstances;
                //update boundries
                if (b[0] < boundries[0]) boundries[0] = b[0];
                if (b[1] < boundries[1]) boundries[1] = b[1];
                if (b[2] > boundries[2]) boundries[2] = b[2];
                if (b[3] > boundries[3]) boundries[3] = b[3];
            }
        }
        //user structure
        public class userInst
        {
            //letter instances
            public Hashtable letters;
            //user name
            public String user;
            //initialize user instrance
            public userInst(String u)
            {
                letters = new Hashtable();
                user = u;
            }
            //add a letter instance
            public void addInst(char l, int[] g, int[] s, int p, int d, int[] b)
            {
                //if char not in letters, add
                if (!letters.Contains(l))
                    letters.Add(l, new letterInst(l));
                //update
                ((letterInst)letters[l]).add(g, s, p, d, b);
            }
            //returns an array of letterInst
            public String[] ToArray()
            {
                ArrayList chars = new ArrayList();
                ArrayList value = new ArrayList(letters.Values);
                foreach (letterInst v in (letterInst[])value.ToArray(typeof(letterInst)))
                    chars.Add(v.letter.ToString());
                return (String[])chars.ToArray(typeof(String));
            }
        }

        public Hashtable loadData;

        private void loadBtn_Click(object sender, EventArgs e)
        {
            FileStream openFile = new FileStream(fileAnalyze.Text, FileMode.Open, FileAccess.Read);
            StreamReader sr = new StreamReader(openFile);
            ArrayList users = new ArrayList();
            String instance = sr.ReadLine();

            loadData = new Hashtable();

            String[] identifiers;
            int[] g = new int[200];
            int[] b = new int[4];
            int[] s = new int[2];
            char l;
            int p;
            int d;

            while (!sr.EndOfStream)
            {
                instance = sr.ReadLine();
                
                identifiers = instance.Split(',');
                if ((identifiers[1].ToCharArray())[0] != ' ')
                {
                    if (!users.Contains(identifiers[247]))
                    {
                        users.Add(identifiers[247]);
                        loadData.Add(identifiers[247], new userInst(identifiers[247]));
                    }
                    l = (identifiers[1].ToCharArray())[0];
                    p = (int)double.Parse(identifiers[13]);
                    d = int.Parse(identifiers[10]);
                    for (int i = 0; i < 200; i++) g[i] = int.Parse(identifiers[46 + i]);
                    s[0] = int.Parse(identifiers[8]);
                    s[1] = int.Parse(identifiers[9]);
                    for (int i = 0; i < 4; i++) b[i] = int.Parse(identifiers[3 + i]);
                    ((userInst)loadData[identifiers[247]]).addInst(l, g, s, p, d, b);
                }
            }
            
            usrSelector.Items.AddRange(users.ToArray());
        }

        private void usrSelector_SelectedValueChanged(object sender, EventArgs e)
        {
            charSelector.Items.Clear();
            charSelector.Items.AddRange(((userInst)loadData[usrSelector.SelectedItem.ToString()]).ToArray());
        }

        private void charSelector_SelectedValueChanged(object sender, EventArgs e)
        {
            Graphics g = dispPanel.CreateGraphics();
            g.Clear(Color.White);
            Font tFont = new Font(FontFamily.GenericSansSerif, 9);
            int gridDim = (dispPanel.Height - 1) / 20;
            letterInst curLetter = (letterInst)((userInst)loadData[usrSelector.SelectedItem.ToString()]).letters[(charSelector.SelectedItem.ToString().ToCharArray())[0]];
            Brush fillBrush;

            int maxGrid = 0;
            for(int i=0; i<200; i++)
                if(curLetter.grid[i]>maxGrid) maxGrid = curLetter.grid[i];

            int color;
            for (int i = 0; i < 10; i++)
            {
                for (int j = 0; j < 20; j++)
                {
                    color = 255 - (int)(255.0 * (double)((double)curLetter.grid[j * 10 + i] / (double)maxGrid));
                    fillBrush = new SolidBrush(Color.FromArgb(color, color, color));
                    g.FillRectangle(fillBrush, i * gridDim + 4, j * gridDim + 7, gridDim, gridDim);
                }
            }

            for (int i = 0; i < 10; i++)
                for (int j = 0; j < 20; j++)
                    g.DrawRectangle(Pens.Black, i * gridDim + 4, j * gridDim + 7, gridDim, gridDim);

            g.DrawString("Average Pressure: " + curLetter.pressures.ToString(), tFont, Brushes.Black, 100, 5);
            g.DrawString("Average Duration: " + curLetter.duration.ToString(), tFont, Brushes.Black, 100, 20);
            g.DrawString("Size: " + (curLetter.size[0]).ToString() + "x" + (curLetter.size[1]).ToString(), tFont, Brushes.Black, 100, 35);
            g.DrawString("Upper Left Bound: " + (curLetter.boundries[0]).ToString() + ", " + (curLetter.boundries[1]).ToString(), tFont, Brushes.Black, 100, 50);
            g.DrawString("Bottom Right Bound: " + (curLetter.boundries[2]).ToString() + ", " + (curLetter.boundries[3]).ToString(), tFont, Brushes.Black, 100, 65);
        }
    }
}